
/////////////////////////////////////////////////////////////////////////////
//
/// @file CnCoeffUtils.h
///
/// @author Tomasz Kapela   @date 2010-04-22
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _CAPD_KAPELA_CNCOEFFUTILS_H_
#define _CAPD_KAPELA_CNCOEFFUTILS_H_
#include <fstream>
#include "capd/diffAlgebra/Jet.hpp"
#include "capd/normalForms/normalForms.h"
//#include "typedefs.h"



template <typename CnCoeffType>
void saveCoeff(const char * name, const CnCoeffType & coeff,
    int prec = capd::TypeTraits<typename CnCoeffType::ScalarType::BoundType>::numberOfDigits() + 1
) {
  if(name){
    std::ofstream out(name);
    if(!out) {
      capd::scon << "Error opening file : " << name << std::endl;
      throw "file open problem";
    }
    coeff.save(out, prec);
    out.close();
  }
}

template <typename CnCoeff>
CnCoeff load(const char * name) {
  std::ifstream input(name);
  if(!input) {
    capd::serr << "Error opening file : " << name << std::endl;
    throw "file open problem";
  }
  CnCoeff coeff;
  input >> coeff;
  input.close();
  return coeff;
}

template <typename CnCoeff>
std::string diamToString(const CnCoeff & coeff, bool verbose = true) {
  std::ostringstream str;
  typedef typename CnCoeff::MatrixType MatrixType;
  typedef typename MatrixType::RowVectorType VectorType;
  VectorType diams(coeff.degree() + 1);
  diams.clear();
  //str << "\ndegree : " << degree() << "    dimension : " << dimension();
  for(int r = 0; r <= coeff.degree(); ++r) {
    if(verbose) {

      if(r)
        str << "\ndiameters of derivatives of order " << r << " :";
      else
        str << "\ndiameters of the value :";
    }
    capd::vectalg::Multipointer mp = coeff.first(r);
    do {
      if(verbose) {

        str << "\n   " << capd::vectalg::Multiindex(coeff.dimension(), mp) << "  : " << maxWidth(coeff(mp));
      }
      if(diams[r] < maxWidth(coeff(mp))) {
        diams[r] = maxWidth(coeff(mp));
      }
      // capd::DVector d(coeff.dimension());
      // capd::vectalg::diameter(coeff(mp),d);
      // str << "\n   "<< capd::vectalg::Multiindex(coeff.dimension(), mp) << "  : " << d ;
    } while(coeff.hasNext(mp));
  }

  str << "\n\n==================\n   the biggest diameter : " << diams;

  str << std::endl;
  return str.str();
}

template <typename CICnCoeff>
std::string complexDiamToString(const CICnCoeff & coeff, bool verbose = true) {

  std::ostringstream str;
  Vector diams(coeff.degree() + 1);

  diams.clear();
  for(int r = 0; r <= coeff.degree(); ++r) {
    capd::vectalg::Multipointer mp = coeff.first(r);
    do {
      for(int i = 0; i < coeff.dimension(); ++i) {
        Float d = (diam(coeff(i, mp).real()) + diam(coeff(i, mp).imag())).rightBound();
        if(diams[r] < d) {
          diams[r] = d;
        }
      }
    } while(coeff.hasNext(mp));
  }

  str << "\n\n==================\n   the biggest diameter : " << diams;

  str << std::endl;
  return str.str();
}

template <typename Output>
void writeNonZeroCoeff(Output & output, const CICnCoeff & normalForm, int order, int precision = 0){
  int oldPrecision = (precision)? std::cout.precision(precision) : std::cout.precision();
  for(int i=0; i<normalForm.dimension(); ++i){
    Multipointer mp = normalForm.first(order);
    do{
      if(normalForm(i,mp)!= IComplex(Float(0.))){
        output << "\n $c_{"<< i << ",(" << Multiindex(normalForm.dimension(),mp) << ")}  =  " << normalForm(i,mp) << "$ \n";
      }
    }while(normalForm.hasNext(mp));
  }
  std::cout.precision(oldPrecision);
}

template <typename CnCoeffType>
std::ostream & printTaylorSeries(std::ostream & str, CnCoeffType  coeff){
  //capd::normalForms::derivativesToSeries(coeff);
  //capd::map::print(str, coeff);
  str.precision(20);
  int dim = coeff.dimension();
  for(int i =0; i<dim; ++i){
   for(int order=0; order <= coeff.degree(); ++order){
    Multipointer mp = coeff.first(order);
    do{
       str << "\n " << (coeff(i,mp));
    }while(coeff.hasNext(mp));
   }
  }

  return str;
}

template<typename MatrixT, capd::autodiff::__size_type DEGREE>
std::ostream & printDiam(
    std::ostream & str, const capd::diffAlgebra::Jet<MatrixT,DEGREE> & coeff,
    int minDegree = 0, int maxDegree  = -1
){
 // typedef typename capd::diffAlgebra::Jet<MatrixT,DEGREE>::Multiindex Multiindex;
  typedef typename capd::diffAlgebra::Jet<MatrixT,DEGREE>::Multipointer Multipointer;
  Float maxD = -1;
  if(maxDegree<0) maxDegree=coeff.degree();
  for(int r=minDegree; r<= maxDegree; ++r){
    if(r)
      str << "Taylor coefficients of order " << r << " :\n";
    else
      str << "value :\n";
    Multipointer mp = coeff.first(r);
    do{
      IVector v = coeff(mp);
      Float d = capd::vectalg::maxWidth(v);
      str << mp << " : "  << d //<< coeff(mp) 
              << "\n";
      maxD = max(maxD, d);
    }while(coeff.hasNext(mp));
  }
  str <<"\n max width = " << maxD << std::endl;
  return str;
}

#endif // _CAPD_KAPELA_CNCOEFFUTILS_H_
