/**
 *  @file h4map.h
 *
 *  @author kapela  @date 2010-11-18
 */

#ifndef H4BASICMAP_H_
#define H4BASICMAP_H_

//#include "capd/poincare/BasicPoincareMap.hpp"
//#include "capd/mpcapdlib.h"
#include "typedefs.h"

#include "capd/normalForms/linearSubstitution.hpp"
#include "Parameters.h"

inline std::string getH4BasicMap(){
//  return "var:x,y,px,py;par:c;fun:px,py,-2*x^3-c*x*y^2,-2*y^3-c*x^2*y;";
//  return "var:x,y,px,py;par:c;fun:px,py,-2*x*x*x-c*x*y*y,-2*y*y*y-c*x*x*y;";
  return "var:x,y,px,py;par:c;fun:px,py,-2*x^2*x-c*x*y^2,-2*y^2*y-c*x^2*y;";
}




template <typename PoincareMapT>
class H4BasicMap{

public:
  typedef PoincareMapT PoincareMap;
  typedef typename PoincareMap::Solver DynamicalSystemType;
  typedef typename DynamicalSystemType::VectorFieldType MapType;
  typedef typename MapType::FunctionType FunctionType;
  
  typedef typename DynamicalSystemType::ScalarType ScalarType;
  typedef typename DynamicalSystemType::VectorType VectorType;
  typedef typename DynamicalSystemType::MatrixType MatrixType;
  typedef typename DynamicalSystemType::StepControlType StepControlType;
  typedef typename MapType::JetType CnCoeffType;
  
  typedef typename capd::poincare::CoordinateSection<MatrixType> SectionType;
  
  H4BasicMap(const ScalarType & c,         ///< parameter 
        const ScalarType & h,         ///< energy level
        int order = 20,               ///< taylor order for solver
        int maxDerivativeOrder = 1,  
        const StepControlType & stepControl = StepControlType()
        
  ) : c(c),
    h(h),
    maxDerivativeOrder(maxDerivativeOrder),
    vfield(getH4BasicMap(),maxDerivativeOrder),
    //section("var:x,y,px,py;fun:y;"),
    section(4,1),
    dynsys(vfield, order, stepControl),
    //dynsys(vfield, order, step),
    poincareMap(dynsys, section, capd::poincare::MinusPlus),
    expansion("var:x,y,px,py;par:h;fun:x,0,px,sqrt(2*h-x*x*x*x-px*px);",maxDerivativeOrder),
    h_energy("var:x,y,px,py;par:c;fun:(px^2+py^2+(x^2)^2+c*x^2*y^2+y^2*y^2)/2;") {
    vfield.setParameter("c", c);
    h_energy.setParameter("c", c);
    expansion.setParameter("h", h);
  }
  

  int dimension() const {
    return 2;
  }
  DynamicalSystemType & getDynSys() { return dynsys; }
  
  void setEnergyLevel(const ScalarType & new_h){
    h = new_h;
    expansion.setParameter("h",h);
  }
  
  VectorType operator()(const VectorType & x){
    poincareMap.turnOnStepControl();
    VectorType z = expand(x);
    capd::sbug << "\n-----\n   x : " << x << "\n    expanded i.c. : " << z << "\n   energy(z) : " << energy(z);

    VectorType pz =  poincareMap(z);
    VectorType px = reduce(pz);
    capd::sbug << "\n   px : " << px << "\n   energy(pz) : " << energy(pz) << "\n-----";
    return px;
  }

  VectorType operator()(const VectorType & x, MatrixType & dPx){

    poincareMap.turnOnStepControl();

    VectorType z = expand(x);
    capd::sbug << "\n-----\n   x : " << x << "\n    expanded i.c. : " << z
        << "\n   energy(z) : " << energy(z);

    MatrixType dPz(4,4);

    capd::sbug <<"\n computing poincare map \n";
    VectorType Pz =  poincareMap(z, dPz);

    dPz = poincareMap.computeDP(Pz, dPz);

    MatrixType expDer = expansion[z];

    VectorType Px = reduce(Pz);
    capd::sbug << "\n   px : " << Px << "\n   energy(Pz) : " << energy(Pz) << "\n-----";

    dPx= reduceDer(dPz*expDer);

    return Px;
  }

  VectorType expand(const VectorType &  x){
    VectorType v(4);
    v[0] = x[0];    v[1] = ScalarType(0.0);
    ScalarType x2 = sqr(x[0]);
    ScalarType py2 = 2*h - sqr(x2)  - sqr(x[1]);
    v[2] = x[1];    v[3] = sqrt(py2);
    return v;
  }

  VectorType reduce(const VectorType & x){
    VectorType v(2);
    v[0] = x[0];
    v[1] = x[2];
    return v;
  }
  MatrixType reduceDer(const MatrixType & d){
    MatrixType r(2,2);
    r[0][0] = d[0][0];   r[0][1] = d[0][2];
    r[1][0] = d[2][0];   r[1][1] = d[2][2];
    return r;
  }

  ScalarType energy(const VectorType &x){
    return h_energy(x);
  }

protected:
  ScalarType c;
  ScalarType h;
  int maxDerivativeOrder;
  MapType vfield;
  SectionType section;
  DynamicalSystemType dynsys;
  PoincareMap poincareMap;
  MapType expansion;
  FunctionType h_energy;

};


//
//
//
//
//template <typename DynSysT>
//class IH4BasicMap{
//public:
//  typedef DynSysT DynamicalSystemType;
//  typedef typename DynamicalSystemType::MapType MapType;
//  typedef typename MapType::FunctionType FunctionType;
//  typedef typename MapType::MatrixType MatrixType;
//  typedef typename MapType::VectorType VectorType;
//  typedef typename MapType::ScalarType ScalarType;
////  typedef capd::poincare::PoincareMapJet<DynamicalSystemType> PoincareMap;
//  typedef capd::poincare::PoincareMap<DynamicalSystemType> PoincareMap;
//
//
//  IH4BasicMap(
//      const ScalarType & c,
//      const ScalarType & h,
//      int order = 20,
//      ScalarType step = 1./1024./128.
//  )
//  : c(c),
//    h(h),
//    vfield(getH4BasicMap()),
//    section("var:x,y,px,py;fun:y;"),
//    dynsys(vfield, order, step, capd::dynsys::ILastTermsStepControl(1, power(10, -capd::TypeTraits<ScalarType>::numberOfDigits()-3), 1. / 1048576.)),
//    poincareMap(dynsys, section, capd::poincare::MinusPlus),
//    expansion("var:x,y,px,py;par:h;fun:x,0,px,sqrt(2*h-x*x*x*x-px*px);"),
//    h_energy("var:x,y,px,py;par:c;fun:(px^2+py^2+(x^2)^2+c*x^2*y^2+y^2*y^2)/2;"){
//    vfield.setParameter("c", c);
//    h_energy.setParameter("c", c);
//    expansion.setParameter("h", h);
//  }
//
//  VectorType operator()(const VectorType & x, int n){
//
//    poincareMap.turnOnStepControl();
//    VectorType z = expand(x);
//    std::cout << "\n-----\n   x0 : " << z << "\n   energy(x0) : " << energy(z);
//
//    capd::dynset::C0Rect2RSet<MatrixType> set(z);
//
//    VectorType px =  poincareMap(set, n);
//    std::cout << "\n   px : " << px << "\n   energy(px) : " << energy(px) << "\n-----";
//    return reduce(px);
//  }
//
//  MatrixType derivative(const VectorType & x, int n){
//    dynsys.setStep(ScalarType(1./64.));
//    poincareMap.turnOffStepControl();
//    VectorType z = expand(x);
//    std::cout << "\n-----\n   x0 : " << z << "\n   energy(x0) : " << energy(z);
//
//    capd::vectalg::EuclLNorm<VectorType, MatrixType> norm;
//    capd::dynset::C1Rect2Set<MatrixType> set(z, norm);
//
//    MatrixType der(4,4);
//    MatrixType expDer = expansion[z];
//    VectorType px =  poincareMap(set, der, n);
//    der = poincareMap.computeDP(px, der);
//
//    std::cout << "\n   px : " << px << "\n   energy(px) : " << energy(px) << "\n-----";
//    std::cout << "\n der : " << der
//         << "\n expDer : " << expDer
//         << "\n ";
//    poincareMap.turnOnStepControl();
//    return reduceDer(der * expDer);
//  }
//
//  VectorType expand(const VectorType &  x){
//    VectorType v(4);
//    v[0] = x[0];    v[1] = ScalarType(0.0);
//    ScalarType x2 = sqr(x[0]);
//    ScalarType py2 = 2*h - sqr(x2)  - sqr(x[1]);
//    v[2] = x[1];    v[3] = sqrt(py2);
//    return v;
//  }
//
//  VectorType reduce(const VectorType & x){
//    VectorType v(2);
//    v[0] = x[0];
//    v[1] = x[2];
//    return v;
//  }
//  MatrixType reduceDer(const MatrixType & d){
//    MatrixType r(2,2);
//    r[0][0] = d[0][0];   r[0][1] = d[0][2];
//    r[1][0] = d[2][0];   r[1][1] = d[2][2];
//    return r;
//  }
//  ScalarType energy(const VectorType &x){
//    return h_energy(x);
//  }
//
//  void test(const VectorType & x){
//    std::cout
//        << "\n x      : " << x
//        << "\n vf     : " <<  vfield(x)
//        << "\n dynsys : " << dynsys(x)
//        << "\n step   : " << dynsys.getStep()
//
//        << std::endl;
//  }
//protected:
//    ScalarType c;
//    ScalarType h;
//    MapType vfield;
//    FunctionType section;
//    DynamicalSystemType dynsys;
//    PoincareMap poincareMap;
//    MapType expansion;
//    FunctionType h_energy;
//
//};
//


#endif /* H4BASICMAP_H_ */
