/**
 *  @file h4map.h
 *
 *  @author kapela  @date 2010-11-18
 */

#ifndef H4MAP_H_
#define H4MAP_H_


#include "H4BasicMap.h"
#include "Parameters.h"
#include "capd/poincare/PoincareMap.hpp"

template <typename PoincareMapT>
class H4Map : public H4BasicMap<PoincareMapT>{


public:
  typedef H4BasicMap<PoincareMapT> BaseType;
  typedef PoincareMapT PoincareMap;
  typedef typename PoincareMap::Solver DynamicalSystemType;
  typedef typename DynamicalSystemType::VectorFieldType MapType;
  typedef typename MapType::FunctionType FunctionType;
  
  typedef typename DynamicalSystemType::ScalarType ScalarType;
  typedef typename DynamicalSystemType::VectorType VectorType;
  typedef typename DynamicalSystemType::MatrixType MatrixType;
  typedef typename DynamicalSystemType::StepControlType StepControlType;
  typedef typename MapType::JetType CnCoeffType;
  
  typedef typename capd::poincare::CoordinateSection<MatrixType> SectionType;
  
  typedef C0Set C0SetType;
  typedef C1Set C1SetType;
protected:
  using BaseType::dynsys;
  using BaseType::section;
  using BaseType::poincareMap;
  using BaseType::vfield;
  using BaseType::expansion;
  using BaseType::c;
  using BaseType::h;
  int orderDer;
  using BaseType::reduce;
  using BaseType::expand;
public:
  H4Map(const ScalarType & c,         ///< parameter 
        const ScalarType & h,         ///< energy level
        int order = 20,               ///< taylor order for solver
        int maxDerivativeOrder = 1,
        const StepControlType & stepControl = StepControlType()       
  ) : BaseType(c, h, order, maxDerivativeOrder, stepControl), 
      orderDer(order) {
  }
  H4Map(const ODEParameters<VectorType> & ode,
        int order = 20,               ///< taylor order for solver
        int maxDerivativeOrder = 1,
        const StepControlType & stepControl = StepControlType() 
  ) : BaseType(ode.param, ode.h, order, maxDerivativeOrder, stepControl),
      orderDer(order){ 
  }
  
  VectorType operator()(const VectorType & x){

    VectorType z = this->expand(x);
    capd::sbug << "\n-----\n   x : " << x << "\n   energy(z) : " << this->energy(z);
    C0SetType set(z);
    VectorType pz =  poincareMap(set);
    VectorType px = this->reduce(pz);
    capd::sbug << "\n   px : " << px << "\n diam : " << diam(px) << "\n   energy(pz) : " << this->energy(pz) << "\n-----";
    return px;
  }

  VectorType operator()(const VectorType & x, MatrixType & dPx){

    VectorType z = this->expand(x);
    capd::sbug << "\n-----\n   x : " << x << "\n   energy(z) : " << this->energy(z);

    MatrixType dPz(4,4);

    capd::sbug <<"\n computing poincare map \n";

    C1SetType set(z);
    VectorType Pz =  poincareMap(set, dPz);
    dPz = poincareMap.computeDP(Pz, dPz);
   
    VectorType Px = reduce(Pz);
    MatrixType expDer = expansion[z];
    dPx= this->reduceDer(dPz*expDer); 
    
    capd::sbug << "\n   px : " << Px << "\n   energy(Pz) : " << this->energy(Pz) << "\n-----";
    
    return Px;
  }

  VectorType operator()(const VectorType & xi, CnCoeffType & dPx, int period = 1){

    int rank = dPx.degree();
    int spaceDimension = 4;
   // int dim = 2;

    //      dynsys.setStep(ScalarType(1./64.));
    //      poincareMap.turnOffStepControl();
    poincareMap.turnOnStepControl();

    VectorType zi = this->expand(xi);
    capd::sbug << "\n-----\n   xi  : " << xi
        << "\n   zi : " << zi
        << "\n   energy(zi) : " << this->energy(zi);

    CnCoeffType dPzi(spaceDimension,rank);
    CnCoeffType expansionCoeff(spaceDimension, rank);
    expansion(zi, expansionCoeff);          // compute Taylor expansion of 'expansion' at zi 
    // typedef capd::dynset::CnRect2<MatrixType> CnSet;
    // capd::vectalg::EuclLNorm<VectorType, MatrixType> norm;
    CnSet set(expansionCoeff);

    capd::sbug <<"\n computing poincare map \n";
    VectorType Pzi(spaceDimension);
    for(int i=0; i<period-1; ++i){
     /// Pzi =  poincareMap(set, dPzi);
      Pzi =  poincareMap(set);
      capd::sbug << "\n iteration " << i << "\n    Pzi = " << Pzi << std::endl;
    }
    Pzi =  poincareMap(set, dPzi);
    dPzi = poincareMap.computeDP(dPzi);

    VectorType Pxi = reduce(Pzi);
    capd::sbug << "\n   P(xi) : " << Pxi << "\n   energy(P(xi)) : " << this->energy(Pzi) ;
    capd::sbug << "\n   P(zi) : " << Pzi << "\n max width : " << maxWidth(Pzi);
    capd::sbug        << "\n-----";

    dPx = this->reduce(dPzi);
   // std::cout << diamToString(dPxi,false);

    return Pxi;
  }
  
  /**
   * We reduce map: R^4 -> R^4 to  R^2 -> R^2
   *
   * @param coeff    series of map with variables (q1,q2,p1,p2)
   * @return         series in variables  (q1,p1)
   */
    CnCoeffType reduce(const CnCoeffType & coeff){
     // We permute coordinates so that coordinates that should remain are on the first places
     VectorType zero(6);
     MapType permutation("var:q1,q2,p1,p2;fun:q1,p1,q2,p2;", this->maxDerivativeOrder);
     MatrixType P = permutation.derivative(zero);
     //   assertContainsZero(P*Transpose(P)-IMatrix::Identity(6));

     CnCoeffType result = capd::normalForms::linearSubstitution(P, coeff, Transpose(P));
     capd::sbug << "\n coefficient before reduction after permutation of variables :" << result.toString(0,4,0,0,3);

     // After permutation multipointers for result and reduced are the same.
     CnCoeffType reduced(2, coeff.degree());
     for(int r = 0; r <= coeff.degree(); ++r) {
       typename CnCoeffType::Multipointer mp = reduced.first(r);
       do {
         for(unsigned int i = 0; i < reduced.dimension(); ++i) {
           reduced(i, mp) = result(i, mp);
         }
       } while(reduced.hasNext(mp));
     }
     capd::sbug << "\n coefficient after reduction :" << reduced.toString();

     return reduced;
   }  
     
};


#endif /* H4MAP_H_ */
