/**
 *  @file existenceProof.h
 *
 *  @author kapela  @date 2011-01-13
 */

#ifndef _KAPELA_EXISTENCE_PROOF_H_
#define _KAPELA_EXISTENCE_PROOF_H_
#include <stdexcept>
#include "ParallelMap.h"

//template <typename MapT>
//typename MapT::VectorType makeExistenceProof(MapT & map, int period, const typename MapT::VectorType & x0, const typename MapT::VectorType & X0, int numberOfIterations = 3){
//
//  typedef typename MapT::VectorType IVector;
//  typedef typename MapT::MatrixType IMatrix;
//
//  ParallelMap<MapT>  pm(map, period);
//
//  IVector fx0 = pm(x0);
//  capd::sbug << "\n f(x0) : " << fx0 << "\n diam f(x0) :" << diam(fx0);
//
//  IVector X = X0;
//  for(int i=0; i<numberOfIterations; ++i){
//    IMatrix dfX(pm.parallelDim(), pm.originalDim());
//    IVector fX = pm(X, dfX);
//    capd::sbug << "\n dF(X) : " << dfX;
//    Interval error;
//    IVector N = pm.computeNewtonOperator(x0, fx0, dfX, error);
//    std::cout << "\n N = " << N << "\n diam(N) : " << diam(N) << "\n X = " << X << "\n N - X " << N-X;
//    if(subset(N, X)){
//      std::cout << "\n PROOF OK!";
//    }else {
//      std::cout << "\n PROOF Failed!";
//    }
//    X = N;
//  }
//  return X;
//}

/**
 * refines rigorous estimates of a zero of the map f
 */
template <typename MapT>
typename MapT::VectorType refineRigorousICParallel(
    MapT & f,                                          ///<  map
    typename MapT::VectorType x0,                      ///<  a point (e.g. approximated zero of f)
    typename MapT::VectorType fx0,                     ///<  value of f computed in x0
    typename MapT::VectorType X,                       ///<  rigorous estimates of the zero of map f
    typename MapT::ScalarType expectedBoxSize,         ///<  expected size of the estimates
    int maxNumberOfIterations = 3                      ///<  maximal number of iterations
){

  if(maxNumberOfIterations <= 0)
    return X;
  
  typedef typename MapT::VectorType IVector;
  typedef typename MapT::MatrixType IMatrix;
  typedef typename MapT::ScalarType ScalarType;

  capd::sout << "\nRefining Rigourous Enclosure." << std::endl;
  //IVector X = X0;

  for(int i=0; i<maxNumberOfIterations; ++i){
    if(!subset(x0,X)){
      capd::sbug << "\n Point x0 is not subset of X. We set x0 = mid(X).";
      x0 = capd::vectalg::midVector(X);
      fx0 = f(x0);
    }
    IMatrix dfX(f.parallelDim(), f.originalDim());
    IVector fX = f(X, dfX);
    capd::sbug << "\n dF(X) : " << dfX;
    Interval error;
    IVector N = f.computeNewtonOperator(x0, fx0, dfX, error);
    capd::sbug << "\n N"<<i <<" = " << N << "\n diam(N" << i <<") : " << diam(N) << std::endl;

    if(! capd::vectalg::intersection(N, X, X)){
      std::cout << "\n\n INTERSECTION IS EMPTY in improveRigorousIC ";
      exit(1);
    }
    ScalarType sizeX = capd::vectalg::maxDiam(X);
    if(sizeX < expectedBoxSize)
      break;

  }
  capd::sout << "\n  refined i.c. : " << X << std::endl;
  return X;
}

/**
 *  It makes proof that given \c map has periodic point of given \c period
 *  in the set with center x0 and radius setSize.
 */
template <typename MapT, typename DVector>
typename MapT::VectorType makeExistenceProof(
    MapT & map,                                     ///< a map
    int period,                                     ///< period (number of iteration of the map)
    const DVector & dx0,                            ///< approximated initial condition
    const typename DVector::ScalarType & setSize,   ///< size of the set with center in x0
    const typename DVector::ScalarType & expectedBoxSize, ///<  expected size of the rigorous estimates
    int maxNumberOfIterations                          ///< number of iterations to improve estimates
){

  capd::sout << "\nMaking existence proof" << std::endl;
  capd::sbug << "\n x0 : " << dx0 << "\n set radius : " << setSize << std::endl;

  typedef typename MapT::VectorType IVector;
  typedef typename MapT::MatrixType IMatrix;
  typedef typename MapT::ScalarType Interval;

  IVector x0(dx0.dimension());
  for(int i=0; i< dx0.dimension(); ++i){
    x0[i] = Interval(dx0[i]);
  }
  IVector X0 = x0 + Interval(-setSize, setSize);

  ParallelMap<MapT>  pm(map, period);

  C0Set x0set(x0);
  IVector fx0 = pm(x0set);
  capd::sbug << "\n f(x0) : " << fx0 << "\n diam f(x0) :" << diam(fx0) << std::endl;

  IVector X = X0;

  IMatrix dfX(pm.parallelDim(), pm.originalDim());
  IVector fX = pm(X, dfX);
  capd::sbug << "\n df(X) : " << dfX << std::endl;

  Interval error;
  IVector N = pm.computeNewtonOperator(x0, fx0, dfX, error);
  capd::sout << "\n N = " << N << "\n\n diam(N) : " << capd::vectalg::convertObject<Vector>(diam(N))
             << "\n\n X = " << X << "\n\n diam(X) : " << capd::vectalg::convertObject<Vector>(diam(X))
             << "\n----------------------------------------" << std::endl;

   //computeNewtonOperator2(x0, fx0, dfX, period);

  if(!subset(N, X)){
    std::cerr << "\n PROOF Failed!" << "\n N-X " << N-X;
    throw std::runtime_error("\n Existence proof failed! ");
  }

  capd::sout << "\n======================\n EXISTENCE PROOF : OK\n======================"<< std::endl;

  X = refineRigorousICParallel(pm, x0, fx0, N, expectedBoxSize, maxNumberOfIterations);

  return X;
}



#endif //_KAPELA_EXISTENCE_PROOF_H_
