/**
 *  @file ioUtils.cpp
 *
 *  @author kapela  @date 2011-01-13
 */

#include "ioUtils.h"
#include <fstream>
#include <stdexcept>

const char
    * default_helpinfo =
        "Usage:"
        "\n  progName [options] dataFile [resultFile] \n\n"
"Options are:\n"
"-h              to display this help information,\n\
-p --precision   precision in mantisa bits,\n"
//"--sout           file name where results will be stored,\n"
//"--scon           file name where console output will be placed,\n"
"--tex            file where data in tex format will be placed \n"
"-q               quiet mode, scon output is not showed on screen.\n"
"-v --verbose     verbose mode - displays additional informations\n"
"--debug         debug mode";




/**
 * Reads file and returns its content without comments.
 * Comments can start with #, ; or %
 *
 * @param name file name
 */
std::string  readAndSkipComments(const char * name){
  char A_c;
  std::ifstream  A_in(name);
  if(! A_in){
    std::cerr << "\n Error opening file " << name << " !" << std::endl;
    throw std::runtime_error("Error opening file !");
  }

  std::ostringstream str;
  do{
  A_in.get(A_c);
  while(A_c == ';' || A_c == '#' || A_c == '%'){
    do{
      A_in.get(A_c);
    }while(A_c!='\n' && !A_in.eof() );
    if (!A_in.eof())
      A_in.get(A_c);
    else
      A_c=' ';
  }
  str << A_c;
//  std::cout << A_c;
  }while(!A_in.eof());
  return str.str();
}

int readCommandLineParameters (
        int argc, char ** argv,
        std::string & dataFile,
        std::string & resultTexFile,
        
        const char * help_info
   ) {
  using namespace capd::auxil;
  int prec = 300;

  arguments a;
  argbreak(a, "h");
  argoblig(a, 0, dataFile);
  arg(a, 0, resultTexFile);
  arg(a, "p", prec);
  arg(a, "-precision", prec);
  arg(a, "-tex", resultTexFile);

  // analyze the command line
  argstreamprepare (a);
  int argresult = a. analyze(argc, argv);
  argstreamset ();
  
  if(argresult < 0) {
    capd::serr << "\n ERROR when reading parameters";
  } else if(argresult > 0) {
    capd::sout << help_info << std::endl;
  } else {
    setPrecision(prec);
    capd::sbug << "\n Precision used : " << prec ;
   
  }
  return argresult;
}

int readCommandLineParametersWithStage(
        int argc, char ** argv,
        std::string & dataFile,
        std::string & outputFile,              
        std::string & resultTexFile,
        ProofStage & stage,
        const char * help_info
   ) {
  using namespace capd::auxil;
  int prec = 300;
  int stageCode = ((int)stage<=0 ? 1 : (int)stage);
//  const char * resultFile = 0;
//  const char * consoleFile = 0;
//  capd::scon.show = true;
//  capd::sout.show = true;
//  capd::sbug.show = false;

  arguments a;
  argbreak(a, "h");
  argoblig(a, 0, dataFile);
  arg(a, 0, outputFile);
  arg(a, "p", prec);
  arg(a, "-precision", prec);
//  arg(a, "-sout", resultFile);
//  arg(a, "-scon", consoleFile);
  arg(a, "-tex", resultTexFile);
  arg(a, "s",stageCode);
  arg(a, "-stage", stageCode);
  
//  argswitch(a, "q", capd::scon.show, false);
//  argswitch(a, "v", capd::sbug.show, true);
//  argswitch(a, "-verbose", capd::sbug.show, true);
//  argswitch(a, "-debug", capd::sbug.show, true);

  // analyze the command line
  argstreamprepare (a);
  int argresult = a. analyze(argc, argv);
  argstreamset ();
  
  if(argresult < 0) {
    capd::serr << "\n ERROR when reading parameters";
  } else if(argresult > 0) {
    if(help_info) 
      capd::sout << help_info << std::endl;
    else 
      capd::sout << default_helpinfo 
                 << "\n -s, --stage   stage of the proof to start with: "
                 << "\n              1-refine 2-existence 3-Cn computation 4-KAM check" << std::endl;           
  } else {
    setPrecision(prec);
    stage = ProofStage(stageCode);
    capd::sbug << "\n Precision used : " << prec ;   
  }
  return argresult;
}

std::shared_ptr<std::ifstream> readConfigFile(
    capd::auxil::ConfigFileReader & ini, 
    const std::string & dataFile,        
    ProofStage stage, // = ProofStage::Refine,
    ICnCoeff * coeff // = nullptr
){
  std::shared_ptr<std::ifstream> file (new std::ifstream(dataFile.c_str()));
  if(!*file){
    std::cerr << "\n Error opening data file : " << dataFile << "\n";
    exit(1);
  }
  ini.setStopTag("raw_data");
  *file >> ini;
  if(stage == ProofStage::KAMProof and coeff != nullptr ){
    *file >> *coeff;
  }
  return file;
}

