/**
 *  @file normalFormComputation.h
 *
 *  @author kapela  @date 2010-10-20
 */

#ifndef NORMALFORMCOMPUTATION_H_
#define NORMALFORMCOMPUTATION_H_


//#include "typedefs.h"
#include "tests.h"
#include "capd/intervals/IComplex.h"
#include "capd/vectalg/Multiindex.h"
#include "capd/normalForms/normalForms.h"
#include "CnCoeffUtils.h"

IComplex complexify(const Interval & x) {
  return IComplex(x, Interval(0.0));
}

/**
 *   Returns rigorous enclosure of eigenvalues of the 2x2 matrix
 *
 *   It throws an exception if we cannot proof that eigenvalues are on the unit circle.
 */

template <typename MatrixType>
CIVector computeEigenvalues2D(const MatrixType & A){

  isSymplectic(A);

  CIVector eigen(2);

  capd::sbug << "\n Compute eigenvalues of A = \n" << A;

  typedef typename MatrixType::ScalarType ScalarType;
  ScalarType a(1.0),
      b(- capd::vectalg::trace(A)) ,
      c(capd::matrixAlgorithms::det(A)) ;

  ScalarType delta = (b*b) - 4*a*c;

  capd::sbug << "\n lambda_1, lambda_2  are solutions of ax^2 + bx + c = 0 where "
      << "\n a = " << a << "\n b= " << b << "\n c = " << c << "\n delta = " << delta;


  assertTrue((delta < 0.0) ,"\nCannot compute eigenvalues");

  ScalarType x = -b / (2*a),
             y = sqrt(-delta)/ (2*a);

  capd::sbug << "\n\n x = " << x << " \n y = " << y;

  assertTrue(((x <1.0) && (x > -1.0 )), "\nCannot compute eigenvalues of the Eight: cannot exclude real case.");

  eigen[0] = IComplex(x, y);
  eigen[1] = IComplex(x, -y);

  capd::sbug << "\n Rigorous eigenvalues : \n  " << eigen;
  capd::sbug << "\n diam x: " << diam(x)  << " y: " << diam(y) ;
  return eigen;
}

/**
 *   Returns rigorous enclosure of eigenvalues of 4x4 matrix
 *   and proofs that they are on the unit circle.
 *
 *   We assume that A is symplectic.
 *   It throws an exception if we cannot proof that eigenvalues are on the unit circle.
 */

template <typename MatrixType>
CIVector computeEigenvalues4D(const MatrixType & A){

  isSymplectic(A);

  //------------------------------------------
  // Computing approximated eigenvalues for checks
  {
  int dim = A.numberOfRows();
   DMatrix Amid(dim, dim);
   std::transform(A.begin(), A.end(), Amid.begin(), midPoint);
   capd::sbug << "\n A = " << A << "\n center of A = " << Amid;
   DVector re(dim), im(dim);
   DMatrix vre(dim, dim), vim(dim, dim);
   capd::alglib::computeEigenvaluesAndEigenvectors(Amid, re, im, vre, vim);

   CIVector approxEigenvalues(dim);
   CIMatrix approxEigenvectors(dim, dim);
   int p[] = {0, 1, 2, 3};
   for(int i=0; i<dim; ++i){
     approxEigenvalues[i] = IComplex(Float(re[p[i]]),Float(im[p[i]]));
     for(int j = 0; j<dim; ++j){
       approxEigenvectors(i+1, j+1) = IComplex(Float(vre(i+1, p[j]+1)), Float(vim(i+1, p[j]+1)));
     }
   }
   capd::sbug << "\n approx eigenvalues " << approxEigenvalues;//capd::alglib::eigenvaluesToString(re, im);
  }
   //----------------------------------------


  CIVector eigen(4);

  typedef typename MatrixType::ScalarType ScalarType;
  ScalarType a(1.0),
      b(- capd::vectalg::trace(A)) ,
      c(capd::vectalg::secondTrace(A)-ScalarType(2.0)) ;

  ScalarType delta = (b*b) - 4*a*c;

  capd::sbug << "\n T1=lambda_1 + lambda_3,   T2= lambda_2 + lambda_4   are solutions of ax^2 + bx + c = 0 where "
      << "\n a = " << a << "\n b= " << b << "\n c = " << c << "\n delta = " << delta;


  assertTrue((delta >= 0.0) ,"\nCannot compute eigenvalues.");

  ScalarType x1 = (-b + sqrt(delta))/ (2*a),
      x2 = (-b - sqrt(delta))/ (2*a);

  capd::sbug << "\n\n T1 = " << x1 << " \n\n T2 = " << x2;

  assertTrue(((x2 <0.0) && (x1 >0.0 )), "\nCannot compute eigenvalues: real parts should have different signs.");
  assertTrue((x1 < 2.0),                "\nCannot compute eigenvalues: cannot exclude real case.");

  x1 /= Float(2.0);
  x2 /= Float(2.0);

  // eigenvalues are on the unit circle
  Interval y1 = sqrt(1 - sqr(x1));
  Interval y2 = sqrt(1 - sqr(x2));

  eigen[0] = IComplex(x1, y1);
  eigen[1] = IComplex(x2, y2);
  eigen[2] = IComplex(x1, -y1);
  eigen[3] = IComplex(x2, -y2);

  capd::sbug << "\n Rigorous eigenvalues : \n  " << eigen;
  capd::sbug << "\n diam x1: " << diam(x1) << " x2: " << diam(x2)
             << " y1: " << diam(y1) << " y2: " << diam(y2);
  return eigen;
}
/**
 * Computes eigenvector corresponding to eigenvalue lambda of a matrix M
 *
 * @param A  = M - lambda * Id
 * @param[out] result  we assume that last coordinate is already set
 */
template <typename MatrixType, typename ResultType>
void computeEigenvector(MatrixType a, ResultType& result) {
  typedef typename MatrixType::ScalarType ScalarType;
  typedef typename ScalarType::value_type ValueType;
  int i, j, k;
  int dimension = a.numberOfRows();
  int *p = new int[dimension];
  for(i = 0; i < dimension; ++i)
    p[i] = i;

  for(j = 0; j < dimension - 1; ++j) {
    ValueType a_max, a_max_temp;
    int i_max, temp;
    a_max = norm(a[p[j]][j]);
    i_max = j;
    temp = p[j];
    for(i = j + 1; i < dimension; ++i) {
      if((a_max_temp = norm(a[p[i]][j])) > a_max) {
        a_max = a_max_temp;
        i_max = i;
      }
    }
    p[j] = p[i_max];
    p[i_max] = temp;

    ScalarType divisor = a[p[j]][j];
    if(isSingular(divisor)) {
      throw std::runtime_error("Gauss elimination: singular matrix\n");
    }

    for(i = j + 1; i < dimension; ++i) {
      ScalarType factor = a[p[i]][j] / a[p[j]][j];
      a[p[i]][j] = capd::TypeTraits<ScalarType>::zero();
      for(k = j + 1; k < dimension; ++k) {
        a[p[i]][k] -= factor * a[p[j]][k];
      }
    }
  }

  if(!isSingular(a[p[dimension - 1]][dimension - 1])) {
    capd::serr << "\n Eigenvector computation : matrix a = (A - lambda * I ) should be singular \n the value = " << a[p[dimension - 1]][dimension - 1] << std::endl;

    throw std::runtime_error( "Eigenvector computation : matrix a = (A - lambda * I ) should be singular\n");
  }

  for(int i = dimension - 2; i >= 0; --i) {
    result[i] = capd::TypeTraits<ScalarType>::zero();
    for(int j = i + 1; j < dimension; ++j)
      result[i] -= a[p[i]][j] * result[j];
    result[i] /= a[p[i]][i];
  }
  delete[] p;
}

IComplex divide(const IComplex & a, const IComplex &b){
  Interval div = sqr(b.real()) + sqr(b.imag());
  Interval re = (a.real()*b.real()+a.imag()*b.imag())/div;
  Interval im = (a.imag()*b.real()-a.real()*b.imag())/div;
  checkIfContainsZero(IComplex(re,im)*b-a);
 // std::cout << "\n a : " << a << "    b : " << b << "\n re : " << re << "  im : " << im << "   div : " << div;
  return IComplex(re, im);
}
//CIVector divide(const capd::vectalg::RowVector<std::complex<capd::intervals::Interval<capd::multiPrec::MpReal, capd::multiPrec::MpReal> >, 0> & a,
//                const IComplex & b){
//  CIVector r(a.dimension());
//  for(int i=0; i< a.dimension(); ++i){
//    r[i] = divide(a[i], b);
//  }
//  return r;
//}
//CIVector divide(const capd::vectalg::ColumnVector<std::complex<capd::intervals::Interval<capd::multiPrec::MpReal, capd::multiPrec::MpReal> >, 0> & a,
//                const IComplex & b){
//  CIVector r(a.dimension());
//  for(int i=0; i< a.dimension(); ++i){
//    r[i] = divide(a[i], b);
//  }
//  return r;
//}
template <typename T>
CIVector divide(const capd::vectalg::RowVector<T, 0> & a,
                const IComplex & b){
  CIVector r(a.dimension());
  for(int i=0; i< a.dimension(); ++i){
    r[i] = divide(a[i], b);
  }
  return r;
}
template <typename T>
CIVector divide(const capd::vectalg::ColumnVector<T, 0> & a,
                const IComplex & b){
  CIVector r(a.dimension());
  for(int i=0; i< a.dimension(); ++i){
    r[i] = divide(a[i], b);
  }
  return r;
}
template <typename MatrixType, typename ResultType>
void complexGauss(MatrixType a, ResultType b, ResultType& result) {
 // typedef typename MatrixType::RowVectorType VectorType;
  typedef typename MatrixType::ScalarType ScalarType;
  typedef typename ScalarType::value_type ValueType;
  int i, j, k;
  int dimension = a.numberOfRows();
  int *p = new int[dimension];
  for(i = 0; i < dimension; ++i)
    p[i] = i;

  for(j = 0; j < dimension - 1; ++j) {
    ValueType a_max, a_max_temp;
    int i_max, temp;
    a_max = norm(a[p[j]][j]);
    i_max = j;
    temp = p[j];
    for(i = j + 1; i < dimension; ++i) {
      if((a_max_temp = norm(a[p[i]][j])) > a_max) {
        a_max = a_max_temp;
        i_max = i;
      }
    }
    p[j] = p[i_max];
    p[i_max] = temp;

    ScalarType divisor = a[p[j]][j];
    if(isSingular(divisor)) {
      throw std::runtime_error("Gauss elimination: singular matrix\n");
    }
    for(i = j + 1; i < dimension; ++i) {
      ScalarType factor = a[p[i]][j] / a[p[j]][j];

      a[p[i]][j] = capd::TypeTraits<ScalarType>::zero();
      b[p[i]] -= factor * b[p[j]];
   
      for(k = j + 1; k < dimension; ++k) {
        a[p[i]][k] -= factor * a[p[j]][k];
      }
    }
  }
  capd::sbug << " \n\n LU zrobione ";
  int start = dimension - 1;
  if(isSingular(a[p[dimension - 1]][dimension - 1])) {
    throw std::runtime_error("Gauss elimination: singular matrix\n");
  }
  for(int i = start; i >= 0; --i) {
    result[i] = b[p[i]];
    for(int j = i + 1; j < dimension; ++j)
      result[i] -= a[p[i]][j] * result[j];
    result[i] = divide(result[i], a[p[i]][i]);
  }
  delete[] p;
}

std::vector<int> getPermutation(const CIVector & eigenvalues, const DVector & re, const DVector & im){
  int dim = eigenvalues.dimension();
  //typedef CIVector::ScalarType Float;

  int index = -1;
  std::vector<int> indices(dim);
  capd::sbug << "\n get Permutation: "
      <<"\n eigenvalues " << eigenvalues
      <<"\n re " << re
      <<"\n im " << im << std::endl;

  for(int i=0; i<dim; ++i){
    Float min_dist = 1.e100;
    for(int j=0; j<dim; ++j){
      Float dist = leftBound(norm((eigenvalues[i] - IComplex(Float(re[j]),Float(im[j])))));
      if(min_dist > dist){
        min_dist = dist;
        index = j;
      }
    }
    if(index<0){
      std::cerr << "\n get permutation : error is to big ! \n";
      exit(1);
    } else {
      indices[i] = index;
    }
  }
  capd::sbug << "\n eigenvalues permutation : ";
  for (int i = 0; i < dim; ++i) {
    capd::sbug << indices[i] << ", ";
  }
  for (int i = 0; i < dim; ++i) {
    for (int j = i+1; j < dim; ++j) {
      if(indices[i]==indices[j]){
        std::cerr << "\n this is not a permutation !";
        exit(2);
      }
    }
  }
  return indices;
}

/**
 * Computes rigorous eigenbasis for given matrix A
 *
 * @param[in]  A           complex matrix
 * @param[out] eigenvalues vector of eigenvalues
 * @param[out] basis       matrix of eigenvectors (as its columns)
 * @param[out] invBasis    inverse of matrix of eigenvectors
 */
//void computeEigenbasis( CIMatrix  A, CIVector & eigenvalues, CIMatrix & basis, CIMatrix & invBasis) {
void computeEigenbasis(const CIMatrix & A, CIVector & eigenvalues, CIMatrix & basis, CIMatrix & invBasis) {

  // We compute double approximations of eigenvalues and eigenvectors
  // This is mainly used for checks
  int dim = A.numberOfRows();

  DMatrix Amid(dim, dim);
  std::transform(A.begin(), A.end(), Amid.begin(), midRePoint);
  capd::sbug << "\n A = " << A << "\n center of A = " << Amid;
  DVector re(dim), im(dim);
  DMatrix vre(dim, dim), vim(dim, dim);
  capd::alglib::computeEigenvaluesAndEigenvectors(Amid, re, im, vre, vim);

  CIVector approxEigenvalues(dim);
  CIMatrix approxEigenvectors(dim, dim);
  std::vector<int> p = getPermutation(eigenvalues, re, im);

  for(int i=0; i<dim; ++i){
    approxEigenvalues[i] = IComplex(Float(re[p[i]]),Float(im[p[i]]));
    for(int j = 0; j<dim; ++j){
      approxEigenvectors(i+1, j+1) = IComplex(Float(vre(i+1, p[j]+1)), Float(vim(i+1, p[j]+1)));
    }
  }
  capd::sbug << "\n approx eigenvalues " << approxEigenvalues;//capd::alglib::eigenvaluesToString(re, im);

  Float tolerance = 1.e-13;
  checkEqual(approxEigenvalues, eigenvalues, tolerance , "approximate eigenvalues almost equal to proven one");

  CIVector zero(dim);
  CIMatrix Id = CIMatrix::Identity(dim);
  // we solve  (A - eigen[i] * Id) x = 0 for x
  for(int i = 0; i < dim/2; ++i) {
    CIMatrix M = (A - eigenvalues[i] * Id);
    CIVector result(dim);
    // We set up last component of eigenvector to be equal to the approximated one (we want also to obtain conjugated eigenvectors)
    result[dim-1] =  approxEigenvectors[dim-1][i];

    computeEigenvector(M, result);
    basis.column(i) = result;
    basis.column(i+dim/2) = capd::vectalg::conjVector(result);
    capd::sbug << "\n eigenvector " << i << " = " << result;
  }

  tolerance = 1.e-10;
  checkEqual(approxEigenvectors, basis, tolerance, "approximate eigenvectors almost equal to rigorous one");
  capd::sbug << "\n basis : " << basis;

// We make basis symmplectic
  CIMatrix B(basis);
  {
    //  IComplex x = B(1,1)*B(3,2)-B(3,1)*B(1,2);

    int n = dim / 2;
    for (int i = 1; i <= n; ++i) {
      IComplex x(Interval(0.0), Interval(0.0));
      /// we compute skew product of B(.,i) and B(.,i+n)
      for (int j = 1; j <= n; ++j) {
        x += B(j, i) * B(j + n, n + i) - B(j + n, i) * B(j, i + n);
      }

      // because B(.,i) and B(.,i+n) are complex conjugate this  value need to be pure imaginary.
      if (x.imag() > 0) {
        IComplex tt = eigenvalues[i - 1];
        eigenvalues[i - 1] = eigenvalues[n + i - 1];
        eigenvalues[n + i - 1] = tt;
        CIVector tmp = B.column(i - 1);
        B.column(i - 1) = B.column(n + i - 1);
        B.column(n + i - 1) = tmp;
        capd::sbug << "\n\n\n interchange of eigenvectors " << i << "  with  " << n + i - 1 << "\n eigenvalues " << eigenvalues << "\n\n\n";
      }
      Interval factor = sqrt(capd::abs(x.imag()));
      IComplex t = IComplex(Interval(0.0), -factor);
      B.column(n + i - 1) = divide(B.column(n + i - 1), t);

      IComplex s = IComplex(factor, Interval(0.0));
      //d = IComplex(Interval(0.0), factor);
      B.column(i - 1) = divide(B.column(i - 1), s);

    }
    // ---------------------- TEST -------------------------
    for (int i = 1; i <= n; ++i) {
      IComplex x(Interval(0.0), Interval(0.0));
      for (int j = 1; j <= n; ++j) {
        x += B(j, i) * B(j + n, n + i) - B(j + n, i) * B(j, i + n);
      }
      capd::sbug << "\n x" << i << " = " << x << std::endl;
      assertContainsZero(x - Interval(1.0));
    }

    assert(isSymplectic(B));
    // --------------------------------------------------------------
  }

  basis = B;

  capd::sbug << "\n\n approx eigenvectors " << capd::alglib::eigenvectorsToString(vre, vim)
             << "\n\n rigorous eigenvectors B = " << basis;

  // we solve equation basis * invBasis = Id  using gauss elimination
  complexGauss(basis, Id, invBasis);

  capd::sbug << "\n\n inverse of B [B1] = " << invBasis;

 // assertContainsZero(tmp);
  // --------------------- TESTS ----------------------------------
  checkEqual(invBasis * basis, Id, tolerance, "B^-1 * B = Id");
  assertContainsZero(invBasis * basis - Id);
  assertContainsZero(basis * invBasis - Id);
  tolerance = 0.0;
  checkEqual(basis * invBasis , Id, tolerance, "B * B^-1 = Id");
  checkEigenvaluesOnDiagonal(eigenvalues, invBasis * A * basis);
  // --------------------------------------------------------------
}

/**
 * Diagonalize C1 part of the poincare map by linear change of coordinates
 *
 * @param coeff          derivatives of the Poincare Map (R^9->R^9) but the revelant functions are  R^4->R^4 and should be on the first positions
 * @param[out] eigenvalues    returns eigenvalues of the eight
 * @return               returns derivatives of Poincare Map R^4 -> R^4 where C1 part is diagonal
 */
CICnCoeff diagonalizeC1(const ICnCoeff & coeff, CIVector & eigenvalues) {

  int dim = coeff.dimension();
  int rank = coeff.degree();

  CICnCoeff complexCoeff(dim, rank);

  for(int r = 0; r <= rank; ++r) {
    CICnCoeff::Multipointer mp = complexCoeff.first(r);
    do {
      for(unsigned int i = 0; i < complexCoeff.dimension(); ++i) {
        complexCoeff(i, mp) = IComplex(coeff(i, mp));
      }
    } while(complexCoeff.hasNext(mp));
  }
  capd::sbug << "\n reduced = " << complexCoeff.toString();

  CIMatrix M = CIMatrix(complexCoeff);

  CIMatrix B(dim, dim);
  CIMatrix B1(dim, dim);

  checkConjugacy(complexCoeff);

  computeEigenbasis(M, eigenvalues, B, B1);

  checkBaseIConjugacy(B);
  capd::sbug << "\n\n eigenbasis = " << B;

  complexCoeff = capd::normalForms::linearSubstitutionPowerSeries(B1, complexCoeff, B);

  checkEigenvaluesOnDiagonal(eigenvalues, static_cast<CIMatrix> (complexCoeff));

  checkRealityCondition(complexCoeff);

  return complexCoeff;
}

IComplex der(const IComplex & coeff, const capd::vectalg::Multiindex & alpha, int index){

  if(alpha[index]==0){
    return capd::TypeTraits<IComplex>::zero();
  } else {
    return Interval(Float(alpha[index])) * coeff ;
  }
}

// computes 1/2 * {F, G}
void poissonBracket(const CICnCoeff & F, int f_degree, int f_index,
                    const CICnCoeff & G, int g_degree, int g_index,
                    CICnCoeff & R, int r_index){
  //int r_degree = f_degree + g_degree -2;
  int dim = F.dimension();
  int dim2 = dim/2;
  typedef capd::vectalg::Multipointer Multipointer;
  typedef capd::vectalg::Multiindex Multiindex;
  Multipointer mpf = F.first(f_degree);
  do{
    Multipointer mpg = G.first(g_degree);
    do{
      Multipointer mpr = capd::vectalg::sumMultipointers(mpf, mpg);
      Multiindex r_a(dim,mpr);
      for(int i = 0; i<dim2; ++i){
        if((r_a[i]!=0) && (r_a[i+dim2]!=0)){
          Multiindex ra2(r_a);
          --ra2[i]; --ra2[i+dim2];
          Multipointer mpr(ra2);
          Multiindex a_f(dim, mpf), a_g(dim, mpg);
          R(r_index,mpr) +=Interval(0.5)*(der(F(f_index,mpf),a_f,i)*der(G(g_index,mpg),a_g,i+dim2)
                           -der(F(f_index,mpf),a_f,i+dim2)
                           *der(G(g_index,mpg),a_g,i));
        }
      }
    }while(G.hasNext(mpg));
  }while(F.hasNext(mpf));
}

template <typename MatrixT>
MatrixT getJ(int dim){
  int dim2=dim/2;
  MatrixT J(dim,dim);
  J.clear();
  for(int i=0; i<dim; ++i){
    J(i+1,  (i+dim2)%dim+1) =  typename MatrixT::ScalarType(Float(((i+dim2)%dim - i)/dim2));
  }
  return J;
}
void derivative(const CICnCoeff & F, int index, int degree, CICnCoeff & der){
  int dim = F.dimension();
  Multipointer mp = der.first(degree-1);
  do {
    Multiindex exponent(dim, mp);
    for(int i=0; i<dim; ++i){
      Multipointer mp2 = capd::vectalg::addIndex(mp, i);
      der(i, mp) = Interval(Float(exponent[i]+1)) * F(index, mp2);
    }
  } while (der.hasNext(mp));
}

CICnCoeff computeNormalFormOrder2(const CICnCoeff & coeffOrig, const CIVector & eigenvalues) {

  capd::sbug << "\n computation of normal form \n";
  CICnCoeff coeff = coeffOrig;

  int dim = coeff.dimension();
  int rank = coeff.degree();

  CICnCoeff N(dim, rank); // change of variables of the form  Id + Quadratic
  for (int i = 0; i < dim; ++i) {
    N(i, i) = capd::TypeTraits<IComplex>::one();
  }
  // N1 is inverse of N
  CICnCoeff N1(N);
  capd::vectalg::Multipointer mp = coeff.first(2);
  capd::sbug << "\n taylor series before normal form computations:\n " << coeff.toString();
  do {
    IComplex divisor = power(eigenvalues, mp);
    for (int i = 0; i < dim; ++i) {
      N(i, mp) = coeff(i, mp) / (divisor - eigenvalues[i]);
      N1(i, mp) = -N(i, mp);
    }
  } while (coeff.hasNext(mp));

  capd::sbug << "\n P " << N.toString();
  CICnCoeff res1(dim, rank),
          res2(dim, rank);

  // We generate Hamiltionian that generates quadratic part
  CICnCoeff H(dim, rank);
  mp = N.first(2);

  int *p = new int[dim]; // = {2, 3, 0, 1};
  int *sign = new int[dim]; // = {1, 1, -1, -1};
  int dim_2 = dim / 2;

  for (int i = 0; i < dim; ++i) {
    p[i] = (i + dim_2) % dim; // {2, 3, 0, 1};
    sign[i] = (i < dim_2) ? 1 : -1; // {1, 1, -1, -1};
  }

  do {
    for (int i = 0; i < dim; ++i) {
      capd::vectalg::Multipointer mp2 = addIndex(mp, p[i]);
      capd::vectalg::Multiindex mi(dim, mp2);
      IComplex c = Interval(sign[i]) * N(i, mp) / IComplex(Interval(mi[p[i]]));
      if (H(0, mp2) == capd::TypeTraits<IComplex>::zero()) {
        H(0, mp2) = c;
      } else {
        assertContainsZero(H(0, mp2) - c);
        assert(capd::intervals::intersection(H(0, mp2), c, H(0, mp2)));
      }
    }
  } while (N.hasNext(mp));

  delete[] p;
  delete[] sign;


  //--------- TESTS ---------------------------------------------
  CICnCoeff Q(dim, rank);
  derivative(H, 0, 3, Q);
  CIMatrix J = getJ<CIMatrix>(dim);
  Q = capd::normalForms::linearSubstitution(J, Q, CIMatrix::Identity(dim)) - N;
  testName("Test: derivative of H is equal Q ");
  testResult(checkIfContainsZero(Q.begin(0, 2), Q.end(0, 2))
          && checkIfContainsZero(Q.begin(1, 2), Q.end(1, 2))
          );
  //---------------------------------------------------------------
  // We compute 3 order terms of coordinate change to make it symplectic up to order 3
  // QH = 1/2*{Q,H}
  CICnCoeff QH(dim, rank);
  for (int i = 0; i < dim; ++i) {
    poissonBracket(N, 2, i, H, 3, 0, QH, i);
  }
  capd::sbug << "\n 1/2*[Q,H] = " << QH.toString(0, dim, 0, 3, 3) << std::endl;
  N = N + QH;

  //  checkSymplecticity(N, "Change N");

  N1 = N1 + QH;

  substitutionPowerSeries(coeffOrig, N, res1, false);
  substitutionPowerSeries(N1, res1, res2, false);

  // --------- Tests -----------------------------------------
  capd::sbug << "\n P " << N.toString() << "\n P1 " << N1.toString()
          << "\n res1 = " << res1.toString()
          << "\n res2 = " << res2.toString();
 
  // checkSymplecticity(N1, "Change N^{-1}");
  testInverseTransformation(N, N1);


  checkEigenvaluesOnDiagonal(eigenvalues, (CIMatrix) res2);
  testName("Test: quadratic part is (contains) zero");
  bool isOk = true;
  for (unsigned int i = 0; i < res2.dimension(); ++i) {
    isOk = checkIfContainsZero(res2.begin(i, 2), res2.end(i, 2)) && isOk;
  }
  testResult(isOk);
  //  checkConjugacy(res2);
  // ----------------------------------------------------------
  return res2;
}

CICnCoeff removeNonResonantTerms(const CICnCoeff & coeffOrig, const CIVector & eigenvalues,  int degree = 3) {


  capd::sbug << "\n---------------------------------------------\n removing non-resonant terms of order " << degree << " (setting them to zero) \n";
  CICnCoeff coeff = coeffOrig;
  int dim = coeff.dimension();
  capd::vectalg::Multipointer mp = coeff.first(degree);
  capd::sbug << "\n original coeffs " << coeff.toString() <<"\n\n resonant terms :\n";
  do {
    IComplex divisor = power(eigenvalues, mp);
    for(int i = 0; i < dim; ++i) {
      IComplex divisor_i = divisor - eigenvalues[i];
      //if(!isSingular(divisor_i)){
      if(capd::abs(divisor_i)>Interval(Float(1.0e-10))){
           coeff(i, mp) = capd::TypeTraits<IComplex>::zero();
      } else {
        capd::sbug << "\n " << i << " | " << capd::vectalg::Multiindex(dim,mp) << " | " << coeff(i,mp);
      }
    }
  } while(coeff.hasNext(mp));

  capd::sbug << "\n\n coeffs after removing terms : " << coeff.toString() << "\n-----------------------------------------------";
  return coeff;
}

/**
 * Computes normal form up to order 3 leaving only resonant terms
 *
 * @return Taylor series up to order 3 in normal form
 */
CICnCoeff computeNormalForm(
    const ICnCoeff & coeffOrig,     ///<  C^3 Jet
    CIVector & eigenvalues    ///<  rigorous estimates of the eigenvalues
){

  CICnCoeff orthogonal = diagonalizeC1(coeffOrig, eigenvalues);

  capd::sbug << "\n orthogonal " << orthogonal << "\n\n";
  capd::sbug << "\n eigenvalues : " << eigenvalues
             << "\n Series after orthogonalization : "<< orthogonal.toString() << complexDiamToString(orthogonal, false);

  CICnCoeff normalForm = computeNormalFormOrder2(orthogonal, eigenvalues);

  normalForm = removeNonResonantTerms(normalForm, eigenvalues, 2);

  capd::sbug << "\n Normal Form (after removing quadratic terms) : \n" << normalForm.toString() << complexDiamToString(normalForm, false);

  CICnCoeff normalForm2 = removeNonResonantTerms(normalForm, eigenvalues, 3);

  capd::sbug << "\n NormalForm2 (after removing non resonant cubic terms) : \n" << normalForm2.toString() << complexDiamToString(normalForm2, false);

  return normalForm2;
}

template <typename Output>
bool checkTwistConditionEight(const CICnCoeff & coeff, const CIVector & eigenvalues, Output & output) {

  IComplex divisor(Interval(2.0) * Interval::pi(), Interval(0.0));  // 2*pi

  Multiindex i(4);
  Multipointer mp;
  int resonantTerms[2][2][4] = {
      {{2,0,1,0}, {1,1,0,1}},
      {{1,1,1,0}, {0,2,0,1}}
  };

  CIMatrix d(2, 2);
  for(int i=0; i<2; ++i){
    for (int j = 0; j < 2; ++j) {
      Multiindex a(4,resonantTerms[i][j]);
      d[i][j] = coeff(i, Multipointer(a))/eigenvalues[i]/divisor;
    }
  }
  IComplex det = d(1, 1) * d(2, 2) - d(1, 2) * d(2, 1);
  capd::sout << "\n D = " << d;
  capd::sout << "\n Final result - torsion \n  det  D = " << det;
  if(!isSingular(det)) {
    capd::sout << "\n Torsion is not singular."
      << "\n==========================="
      << "\n  EIGHT IS KAM STABLE!!!\n"
      << "\n==========================="
      << std::endl;
    output.precision(3);
    output <<"\n D = $\\left(\\begin{array}{cc} \n" 
            << d(1, 1) <<"&" << d(1, 2)   <<"\\\\\n" << d(2, 1) <<"&" << d(2, 2)
            <<"\n\\end{array}\\right)$\n";
    output <<"\n\n det(D) = $" << det << "$\n";
    return true;
  }
  return false;
}

template <typename Output>
bool checkTwistCondition2D(const CICnCoeff & coeff, const CIVector & eigenvalues, Output & output) {

  //IComplex divisor(Interval(1.0), Interval(0.0));  // 2*pi*i
    IComplex divisor(Interval(2.0) * Interval::pi(), Interval(0.0));  // 2*pi

  capd::sout << "\n Checking Twist Condition \n";
  IComplex d;
  int indx[] = {2, 1};
  Multiindex a(2,indx);
  d = coeff(0, Multipointer(a))/eigenvalues[0]/divisor;
  capd::sout << "\n torsion without factor 2pi : " << coeff(0, Multipointer(a))/eigenvalues[0];
  IComplex det = d;
  capd::sbug << "\n D = " << d;
  capd::sout << "\n\nFinal result :\n  torsion = " << det;
  if(!isSingular(det)) {
    capd::sout << "\n\n Torsion is not singular!\n"
        << "\n==========================="
        << "\n  ORBIT IS KAM STABLE !!!"
        << "\n==========================="
        << std::endl;
    output << "\n\n D = $\\left(" << d << "\\right)$"; 
    output <<"\n\n det(D) = $" << d << "$\n";
    return true;
  }
  return false;
}

#endif /* NORMALFORMCOMPUTATION_H_ */
