/////////////////////////////////////////////////////////////////////////////
//
/// @file tests.h
///
///  Various test for proofs. They tests correctness of results.
///
/// @author Tomasz Kapela   @date 2010-03-17
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _CAPD_KAPELA_TESTS_H_
#define _CAPD_KAPELA_TESTS_H_
#include "typedefs.h"
#include "capd/auxil/logger_deprecated.h"
#include "capd/matrixAlgorithms/capd2alglib.h"

inline double midPoint(const Interval & i) {
  return toDouble((i.leftBound() + i.rightBound()) / 2);
}
inline double midRePoint(const IComplex & i) {
  return toDouble((i.real().leftBound() + i.real().rightBound()) / 2);
}


inline void testName(const std::string & name) {
  capd::sbug << "\n " << name << " : ";
}

inline bool testResult(bool result) {
  if(result)
    capd::sbug << "\n  [PASSED]";
  else{
    capd::sbug << "\n  [FAILED]";
    throw std::runtime_error("One of tests failed!!!");
  }
  return result;
}

bool checkIfContainsZero(const Interval &x){
  return x.contains(Float(0.0));
}
bool checkIfContainsZero(const IComplex &x){
  return (real(x).contains(Float(0.0)) && imag(x).contains(Float(0.0)));
}

template <typename Iterator>
bool checkIfContainsZero(const Iterator & begin, const Iterator & end){
   for(Iterator it = begin; it!=end; ++it){
     if(!(capd::abs(*it).contains(0.0))){
       capd::serr << "\n ERROR coefficient : " << *it << " do not contains zero !";
       return false;
     }
   }
   return true;
}

template <typename Object>
bool checkIfReal(const Object & object){
   bool isOK = true;
   for(typename Object::const_iterator it = object.begin(); it!=object.end(); ++it){
     if(!(it->imag().contains(0.0))){
       capd::serr << "\n ERROR coefficient : " << *it << " is not a REAL number !";
       isOK = false;
     }
   }
   return isOK;
}
template <>
bool checkIfReal(const IComplex &x){
  if(x.imag().contains(Float(0.0)))
    return true;
  else
    return false;
}

template <typename Object>
void assertContainsZero(const Object & object){
  bool isOk =  checkIfContainsZero(object.begin(), object.end());
  if(! isOk)
    throw std::runtime_error(" All coordinates should contain zero !");
}

template<>
void assertContainsZero(const Interval & x){
  if(!(x.contains(0.0))){
    capd::serr << "\n Interval " << x << " has to contain zero" << std::endl;
    throw std::runtime_error("Interval should contains zero");
  }
}

template<>
void assertContainsZero(const IComplex & x){
  if(!(real(x).contains(0.0)) || !(imag(x).contains(0.0))){
    capd::serr << "\n IComplex " << x << " has to contain zero" << std::endl;
    throw std::runtime_error("IComplex should contains zero");
  }
}

void assertTrue(bool condition, std::string message="assertion failed!"){
  if(!condition)
    throw std::runtime_error(message);
}
void assertFalse(bool condition, std::string message="assertion failed!"){
  if(condition)
    throw std::runtime_error(message);
}

/**
 * checks if all coefficients of given components of coeff contain zero
 * @param coeff coefficient container
 * @param start first component index
 * @param stop  last component index
 * @return true if all coefficients contain zero, false otherwise
 */
template <typename CnCoeff>
bool containsZero(const CnCoeff & coeff, int start, int stop) {
  for(int i = start; i <= stop; ++i) {
    if(!checkIfContainsZero(coeff.begin(i), coeff.end(i)))
      return false;
  }
  return true;
}
/**
 * Checks if functions from start to stop which derivatives are stored in coeff
 * do not depend on variables with indices [firstVariable,..,lastVariable]
 * If check fail an exception is thrown.
 */
template <typename CnCoeff>
void assertDoNotDepend(const CnCoeff & coeff, int start, int stop, int firstVariable, int lastVariable){
  for(int level=1; level < coeff.degree(); ++level){
    typename CnCoeff::Multipointer mp = coeff.first(level);
    for(int i = start; i<=stop; ++i){
      for(int j=0; j<mp.size(); ++j){
        if((mp[j]>=firstVariable) && (mp[j]<=lastVariable)){
          assertContainsZero(coeff(i,mp));
          break;
        }
      }
    }
  }
}

// checks if w is subset of z
inline bool subset(const IComplex & w, const IComplex & z) {
  return subset(w.real(), z.real()) && subset(w.imag(), z.imag());
}
inline bool checkEigenvaluesOnDiagonal(const CIVector & eigenvalues, const CIMatrix & A) {
  testName("Test: eigenvalues on diagonal");
  capd::slog2 << "\n   eigenvalues : " << eigenvalues << "\n  matrix : " << A;
  bool isOK = true;
  for(unsigned int i = 0; i < eigenvalues.dimension(); ++i) {
    for(unsigned int j = 0; j < eigenvalues.dimension(); ++j){
      if(i==j){
        if(!subset(eigenvalues[i], A(i + 1, i + 1))) {
          isOK = false;
          capd::serr << "\n NOT A SUBSET : \n   " << eigenvalues[i] << "\n   " << A(i + 1, i + 1);
        }
      }else{
        if(!(capd::abs(A(i+1,j+1)).contains(0.0))){
          isOK = false;
          capd::serr << "\n entry ( " << i <<"," << j << " ) =   " << A(i + 1, i + 1) << " SHOULD CONTAINS ZERO! ";
        }
      }
    }
  }
  testResult(isOK);
  return isOK;
}

inline void checkEigenvalues(const ICnCoeff & coeff) {
  int dim = coeff.dimension();
  IMatrix A = IMatrix(coeff);
  DMatrix B(dim, dim);
  std::transform(A.begin(), A.end(), B.begin(), midPoint);
  DVector re(dim), im(dim);
  capd::alglib::computeEigenvalues(B, re, im);
  capd::sbug << "\n B " << B;
  capd::sbug << "\n eigenvalues of C1 part: " << capd::alglib::eigenvaluesToString(re, im);
}

template <typename Container>
inline void checkEqual(const Container & a, const Container & b, const Float & tolerance , const std::string & name = ""){
  testName("Test : "+name);
  typename Container::const_iterator  ita = a.begin(),
                                e = a.end(),
                                itb = b.begin();
  bool equal = true;
  while(ita != e){
    if(capd::abs(*ita - *itb) > tolerance){
      equal = false;
      capd::serr << "\n  NOT EQUAL : \n     " << *ita << "\n  != " << *itb << "\n  diff " << capd::abs(*ita - *itb) ;
    }
    ++ita;  ++itb;
  }
  testResult(equal);
}

template <typename MapT>
class FirstIntegrals{
private:
  FirstIntegrals(){
  }
  FirstIntegrals(const FirstIntegrals &){
  }
public:
  typedef MapT MapType;

  static MapType & getInstance(){
     if(!S_firstIntegrals){
       std::string integrals = std::string("var:x1,y1,px1,py1,x2,y2,px2,py2;fun:")
                     + "(px1^2+py1^2)/2+(px2^2+py2^2)/2+((px1+px2)^2+(py1+py2)^2)/2"
                     +"-((x1-x2)^2+(y1-y2)^2)^(-0.5)-((x1+2*x2)^2+(y1+2*y2)^2)^(-0.5)-((2*x1+x2)^2+(2*y1+y2)^2)^(-0.5)," //energy
                     + "x1*py1-y1*px1+x2*py2-y2*px2+(-x1-x2)*(-py1-py2)-(-y1-y2)*(-px1-px2)," // angular momentum
                     + "0,0,0,0,0,0;"; // artificial zeroes to make function R^n -> R^n
       S_firstIntegrals = new MapType(integrals);
     }
     return (*S_firstIntegrals);
  }
protected:
  static MapType * S_firstIntegrals;
};
template <typename MapT>
MapT * FirstIntegrals<MapT>::S_firstIntegrals = 0;

/// We check energy and angular momentum preservation
/// series of J(x0) and J(solution) up to given order should be the same
inline bool checkFirstIntegralsPreservation(const IVector & x0, const ICnCoeff & solution, int order, int dispRank = 3) {
  testName(" First integrals preservation ");
//  std::string integrals = std::string("var:x1,y1,px1,py1,x2,y2,px2,py2;fun:")
//               + "(px1^2+py1^2)/2+(px2^2+py2^2)/2+((px1+px2)^2+(py1+py2)^2)/2"
//               +"-((x1-x2)^2+(y1-y2)^2)^(-0.5)-((x1+2*x2)^2+(y1+2*y2)^2)^(-0.5)-((2*x1+x2)^2+(2*y1+y2)^2)^(-0.5)," //energy
//               + "x1*py1-y1*px1+x2*py2-y2*px2+(-x1-x2)*(-py1-py2)-(-y1-y2)*(-px1-px2)," // angular momentum
//               + "0,0,0,0,0,0;"; // artificial zeroes to make function R^n -> R^n
//  //solution() = x0;
  IVector x1 = solution();
  IMap J = FirstIntegrals<IMap>::getInstance();
  ICnCoeff Jx0(8,order);
  ICnCoeff Jx1(8,order);
  ICnCoeff Jsolution(8,order);
  J(x0, Jx0);
  J(x1, Jx1);

  substitutionPowerSeries(Jx1, solution, Jsolution, false);
  //Jsolution() = x0;
  ICnCoeff difference = Jsolution - Jx0;

  capd::sbug
      << "\n x0 " << x0
      << "\n x1 " << x1
      << "\n solution " << solution.toString(0,7,0,0,dispRank)
      << "\n der " << (IMatrix)solution
      << "\n Jx0 " << Jx0.toString(0,1,0,0,dispRank)
      << "\n Jsolution " << Jsolution.toString(0,1,0,0,dispRank)
      << "\n difference " << difference.toString(0,1,0,0,dispRank);
  return testResult(containsZero(difference));
}

/// We check that energy and angular momentum preservation are fixed
/// so all derivatives should be 0
inline bool checkFixedFirstIntegrals( const ICnCoeff & solution, int order) {
  testName("Test: first integrals fixed");
  std::string integrals = std::string("var:x1,y1,px1,py1,x2,y2,px2,py2;fun:")
               + "(px1^2+py1^2)/2+(px2^2+py2^2)/2+((px1+px2)^2+(py1+py2)^2)/2"
               +"-((x1-x2)^2+(y1-y2)^2)^(-0.5)-((x1+2*x2)^2+(y1+2*y2)^2)^(-0.5)-((2*x1+x2)^2+(2*y1+y2)^2)^(-0.5)," //energy
               + "x1*py1-y1*px1+x2*py2-y2*px2+(-x1-x2)*(-py1-py2)-(-y1-y2)*(-px1-px2)," // angular momentum
               + "0,0,0,0,0,0;"; // artificial zeroes to make function R^n -> R^n
  //solution() = x0;
  IVector x1 = solution();
  IMap J = integrals;
  ICnCoeff Jx1(8,order);
  ICnCoeff Jsolution(8,order);
  J(x1, Jx1);

  substitutionPowerSeries(Jx1, solution, Jsolution, false);
  IVector zero(solution.dimension());
  zero.clear();
  Jsolution() = zero;

  return testResult(containsZero(Jsolution));
}

/**
 * Checks if two vectors  are conjugate.
 *
 * We assume than variables are (z1, z2, w1, w2) and that z1=\bar{w1}, z2 = \bar{w2}
 * @param coeff
 */
template <typename VectorType>
inline void checkVectorsConjugacy(const VectorType & x, const VectorType & y){
  testName("Test: check conjugacy of two vectors");
  bool isOK = true;
  for(int i=0; i<x.dimension(); ++i){
    if(!checkIfContainsZero(x[i]-conj(y[i]))){
      isOK = false;
      capd::serr << "\n ERROR coefficients : " << x[i] << " and " << y[i] << " are not conjugated !";

    }
  }
  testResult(isOK);
}

/**
 * Checks if two vectors  are conjugate.
 *
 * We assume than variables are (z1, z2, w1, w2) and that z1= i \bar{w1}, z2 = i \bar{w2}
 * @param coeff
 */
template <typename VectorType>
inline bool checkVectorsIConjugacy(const VectorType & x, const VectorType & y){
  testName("Test: check i conjugacy of two vectors");
  bool isOK = true;
  IComplex I(Interval(0.0), Interval(1.0));
  for(int i=0; i<x.dimension(); ++i){
    if(!checkIfContainsZero(x[i]-I*conj(y[i]))){
      isOK = false;
      capd::serr << "\n ERROR coefficients : " << x[i] << " and " << y[i] << " are not i conjugated !";

    }
  }
  return testResult(isOK);
}

template <typename MatrixType>
inline bool  checkBaseIConjugacy(const MatrixType & B){
  testName("Test: check i conjugacy of vectors in base");
  bool isOK = true;
  int dim = B.numberOfColumns() /2;
  for(int i=0; i<dim; ++i){
    if(!checkVectorsIConjugacy(B.column(i), B.column(dim+i))){
      isOK = false;
      capd::serr << "\n ERROR vectors : " << B.column(i) << " and " << B.column(dim+i) << " are not i conjugated !";
    }
  }
  return testResult(isOK);
}

/**
 * Checks if values for conjugate variables are conjugate.
 *
 * We assume than variables are (z1, z2, w1, w2) and that z1=\bar{w1}, z2 = \bar{w2}
 * @param coeff
 */
inline void checkConjugacy(const CICnCoeff & coeff){
  testName("Test: check conjugacy of values corresponding to conjugated variables");
  bool isOK = true;
  int dim = coeff.dimension();
  int dim2 = dim/2;
  //CIMatrix P(4,4);
  //P(1,3) = P(3,1) = P(2,4) =P(4,2) =IComplex(Float(1.0));

  for(int level = 1; level<=3; ++level){
    CICnCoeff::Multipointer mp = coeff.first(level);
    do{
      CICnCoeff::Multiindex index(dim, mp);
      CICnCoeff::Multiindex iconj(dim);
      for(int i=0; i<dim2; ++i){
        iconj[i] = index[dim2+i];
        iconj[dim2+i] = index[i];
      }
      CICnCoeff::Multipointer mpConj(iconj);
      for(int i=0; i<dim2; ++i){
        if(!checkIfReal(coeff(i,mp)+coeff(dim2+i,mpConj))){
          capd::serr << "\n mp = " << mp << "     mpConj = " << mpConj
              << "\n index = " << index << "   index conj. " << iconj
              << "\n coeff(mp)    " << coeff(mp) << "\n coeff(mpConj)" << coeff(mpConj);
          isOK = false;
        }
      }
    }while(coeff.hasNext(mp));
  }
  testResult(isOK);
}

/**
 * Checks if values for conjugate variables satisfy reality condition.
 *
 * We assume than variables are (z1, z2, w1, w2) and that z1=i \bar{w1}, z2 = i \bar{w2}
 * @param coeff
 *
 * TODO: do not work properly !
 */
inline void checkRealityCondition(const CICnCoeff & coeff){
  testName("Test: check reality condition for conjugated variables");
  bool isOK = true;
  int dim = coeff.dimension();
  int dim2 = dim/2;
  //CIMatrix P(4,4);
  //P(1,3) = P(3,1) = P(2,4) =P(4,2) =IComplex(Float(1.0));
  IComplex I(Interval(0.0), Interval(1.0));
  IComplex In =I;
  for(int level = 1; level<=3; ++level){
    CICnCoeff::Multipointer mp = coeff.first(level);
    do{
      CICnCoeff::Multiindex index(dim, mp);
      CICnCoeff::Multiindex iconj(dim);
      int sum=0;
      for(int i=0; i<dim2; ++i){
        iconj[i] = index[dim2+i];
        iconj[dim2+i] = index[i];
        sum += index[i];
      }
      CICnCoeff::Multipointer mpConj(iconj);
      for(int i=0; i<dim2; ++i){
        IComplex c1 = I * conj(coeff(i,mp)),
                 c2 = In * coeff(dim2+i,mpConj);

        if(!(checkIfContainsZero(c1 - c2)  || checkIfContainsZero(c1 + c2)
            )){
          capd::serr << "\n mp = " << mp << "     mpConj = " << mpConj << " i= " << i
              << "\n index = " << index << "   index conj. " << iconj
              << "\n c1  " << c1
              << "\n c2  " << c2
              << "\n -   " << c1 - c2
              << "\n +   " << c1 + c2
              << "\n In  " << In;
          isOK = false;
        }
      }

    }while(coeff.hasNext(mp));
    In *= I;
  }
  testResult(isOK);
}

template <typename Matrix>
bool isSymplectic(const Matrix & A){
  testName("Test: Matrix is symplectic");
  int dim = A.numberOfRows();
  int dim2 = dim/2;
  Matrix J(dim, dim);
  J.clear();
  for(int i=0; i<dim; ++i){
    J(i+1,  (i+dim2)%dim+1) =  typename Matrix::ScalarType(Float(((i+dim2)%dim - i)/dim2));
  }
  capd::sbug << "\n A =" << A ;
  capd::sbug << "\n J =" << J ;
  Matrix B = Transpose(A)*J*A-J;
  capd::sbug << "\n A^T*J*A - J =" << B ;
 return testResult(checkIfContainsZero(B.begin(),B.end()));
}

template <typename CoeffType>
CoeffType transpose(CoeffType A){
  typename CoeffType::iterator i1 = A.begin(1), i2 = A.begin(2), e = A.end(1);

  while(i1!=e){
    std::swap(*i1, *i2);
    i1++;  i2++;
  }
  return A;
}
template <typename CoeffType>
CoeffType multiply(const CoeffType & A, const CoeffType & B){
  int dim = A.dimension();
  const int dim2 = 2;
  int rank = A.degree();
  CoeffType res(dim, rank);
  int m[dim2][dim2] ={{0, 1}, {2, 3}};

  for(int degree=0; degree<=rank; ++degree){
    for(int a=0; a<=degree; ++a){
      capd::vectalg::Multipointer mpa = A.first(a);
      do{
        capd::vectalg::Multipointer mpb = B.first(degree-a);
        do{
          capd::vectalg::Multipointer mpc = sumMultipointers(mpa, mpb);
          for(int i=0; i<dim2; ++i){
            for(int j=0; j<dim2; ++j){
              for(int k=0; k<dim2; ++k){
                res(m[i][j], mpc) += A(m[i][k], mpa) * B(m[k][j], mpb);
              }
            }
          }
        }while(B.hasNext(mpb));
      }while(A.hasNext(mpa));
    }
  }
  return res;
}
template <typename CoeffType>
bool checkSymplecticity(const CoeffType & A, const std::string & name = "Taylor series"){
  testName("Test: " + name + " is symplectic");
  int rank = A.degree();
  int dim = A.dimension();
  if(dim != 4)
    throw std::runtime_error("checkSymplecticity implemented only for dimension 4.");
  //int dim2 = dim/2;
  CoeffType Q_q(4, rank-1), Q_p(4, rank-1),
            P_q(4, rank-1), P_p(4, rank-1);
  {
  int r[] = {0, 1, 0, 1};
  int k[] = {0, 0, 1, 1};
  for(int level=0; level<rank; ++level){
    capd::vectalg::Multipointer mp = Q_q.first(level);
    do{
    for(int i =0; i<dim; ++i){
      capd::vectalg::Multipointer mp2 = capd::vectalg::addIndex(mp, r[i]);
      capd::vectalg::Multiindex index(4, mp2);
    //  capd::sout << index;
      Q_q(i,mp) = A(k[i],mp2) * typename CoeffType::ScalarType(Float(index[r[i]]));
    }
    }while(Q_q.hasNext(mp));
  }
  }
  {
    int r[] = {2, 3, 2, 3};
    int k[] = {0, 0, 1, 1};
    for(int level=0; level<rank; ++level){
      capd::vectalg::Multipointer mp = Q_p.first(level);
      do{
        for(int i =0; i<dim; ++i){

        capd::vectalg::Multipointer mp2 = capd::vectalg::addIndex(mp, r[i]);
        capd::vectalg::Multiindex index(4, mp2);
        Q_p(i,mp) = A(k[i],mp2) * typename CoeffType::ScalarType(Float(index[r[i]]));
      }
      }while(Q_p.hasNext(mp));
    }
  }
  {
    int r[] = {0, 1, 0, 1};
    int k[] = {2, 2, 3, 3};
    for(int level=0; level<rank; ++level){
      capd::vectalg::Multipointer mp = P_q.first(level);
      do{
        for(int i =0; i<dim; ++i){

        capd::vectalg::Multipointer mp2 = capd::vectalg::addIndex(mp, r[i]);
        capd::vectalg::Multiindex index(4, mp2);
        P_q(i,mp) = A(k[i],mp2) * typename CoeffType::ScalarType(Float(index[r[i]]));
      }
      }while(P_q.hasNext(mp));
    }
  }
  {
    int r[] = {2, 3, 2, 3};
    int k[] = {2, 2, 3, 3};
    for(int level=0; level<rank; ++level){
      capd::vectalg::Multipointer mp = P_p.first(level);
      do{
      for(int i =0; i<dim; ++i){
        capd::vectalg::Multipointer mp2 = capd::vectalg::addIndex(mp, r[i]);
        capd::vectalg::Multiindex index(4, mp2);
        P_p(i,mp) = A(k[i],mp2) * typename CoeffType::ScalarType(Float(index[r[i]]));
      }
      }while(P_p.hasNext(mp));
    }
  }

  CoeffType D11 = multiply(transpose(Q_q), P_q) - multiply(transpose(P_q), Q_q);

  CoeffType D12 = multiply(transpose(Q_q), P_p) - multiply(transpose(P_q), Q_p);
  D12(0) -= Float(1.0); D12(3) -= Float(1.0);

  CoeffType D21 = multiply(transpose(Q_p), P_q) - multiply(transpose(P_p), Q_q);
  D21(0) += Float(1.0); D21(3) += Float(1.0);

  CoeffType D22 = multiply(transpose(Q_p), P_p) - multiply(transpose(P_p), Q_p);

  capd::sbug << "\n A " << A.toString();
  capd::sbug << "\n Q_q " << Q_q.toString() << "\n P_q " << P_q.toString()
             << "\n Q_p " << Q_p.toString() << "\n P_p " << P_p.toString()
             << "\n P_p^T "<< transpose(P_p).toString();
  capd::sbug <<"\n D12 = "<< D11.toString()
      << "\n A = " << typename CoeffType::MatrixType(A);
  //isSymplectic(typename CoeffType::MatrixType(A));
  return testResult(
                    checkIfContainsZero(D11.begin(),D11.end()) &&
                     checkIfContainsZero(D12.begin(),D12.end()) &&
                  //  checkIfContainsZero(D21.begin(),D21.end()) &&
                    checkIfContainsZero(D22.begin(),D22.end())
      );
}

template <typename Matrix>
bool isSymplectic2(const Matrix & A){
  testName("Test: Matrix is symplectic");
  int dim = A.numberOfRows();
  Matrix J(dim, dim);
  J.clear();
  for(int i=0; i<dim; i += 4){
    J(i+1, i+3) = 1.0;
    J(i+2, i+4) = 1.0;
    J(i+3, i+1) = -1.0;
    J(i+4, i+2) = -1.0;
  }
  capd::slog2 << "\n J =" << J ;
  Matrix B = A*J*Transpose(A)-J;
  capd::sbug << "\n B =" << B ;
 return testResult(checkIfContainsZero(B.begin(),B.end()));
}
/// tests if P1 is inverse transformation of P.
template <typename CoeffType>
bool testInverseTransformation(const CoeffType & P, const CoeffType & P1){
  testName("Test: Transformations are inverse of each other. ");
  typedef typename CoeffType::ScalarType ScalarType;

  assert(P.dimension()==P1.dimension());
  assert(P.degree()==P1.degree());
  CoeffType R(P.dimension(), P.degree());
  substitutionPowerSeries(P, P1, R, false);
  for(int i=0; i<P.dimension(); ++i){
    R(i,i) -= ScalarType(Float(1.0));
  }
  bool result = checkIfContainsZero(R.begin(), R.end());
  substitutionPowerSeries(P1, P, R, false);
  for(int i=0; i<P.dimension(); ++i){
    R(i,i) -= ScalarType(Float(1.0));
  }
  return testResult(result && checkIfContainsZero(R.begin(), R.end()));

}
template <typename Matrix>
bool isSkewSymmetric(const Matrix & A){
  testName("Test: Matrix is skew-symmetric");
  Matrix B = Transpose(A)-A;
  capd::sbug << "\n B =" << B ;
 return testResult(checkIfContainsZero(B.begin(),B.end()));
}

#endif // _CAPD_KAPELA_TESTS_H_
